#!/usr/bin/env python
# coding: utf-8

from music_dl.MusicDL import MusicDL
from music_dl.core.args import parse_args

__version__ = '0.2.1'
__license__ = 'MIT'
__author__ = 'Gumob'
__author_email__ = 'hello@gumob.com'
__url__ = 'http://github.com/gumob/music-dl'
__copyright__ = 'Copyright (C) 2018 Gumob'
__all__ = ['main', 'MusicDL']


def main():
    args = parse_args()

    """ Execute download """

    mdl = MusicDL(
        download_url=args.url,
        working_dir=args.dir,
        audio_codec=args.codec,
        audio_bitrate=args.bitrate,
        playlist_start=int(args.playlist_start),
        playlist_end=int(args.playlist_end),
        no_artwork=args.no_artwork,
        no_album_title=args.no_album_title,
        no_album_artist=args.no_album_artist,
        no_track_number=args.no_track_number,
        no_composer=args.no_composer,
        no_compilation=args.no_compilation,
        open_dir=args.open_dir,
        # clear_cache=args.clear_cache,
        verbose=args.verbose
    )
    mdl.download()



#!/usr/bin/env python
# coding: utf-8


def padded_num(num=-1, max_num=1):
    """
    The function that convert int to zero padded string

    :param int num: Number to be converted to string
    :param int max_num: Numbers to determine digits
    :return:
    """

    digit = len(str(max_num))
    if num >= 0:
        return str(num).zfill(digit)
    else:
        return '?' * digit


def contains_at_least(l1, l2):
    """
    Checking if two lists share at least one element
    https://stackoverflow.com/questions/24270711/checking-if-two-lists-share-at-least-one-element

    :param list l1: Array to compare
    :param list l2: Array to compare
    :return: bool result: If two lists share at least one element, return true
    """

    return any(x in l1 for x in l2)



# coding:utf-8


class YoutubeDownloadMusicException(Exception):
    """Raised when youtube-dl module throws an exception"""

    def __init__(self, message):
        self.msg = message


class DirectoryException(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class PlaylistParameterException(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class PlaylistPreprocessException(Exception):
    def __init__(self, message, data):
        super().__init__(message)
        self.message = message
        self.data = data


class PlaylistDownloadException(Exception):
    def __init__(self, message, data):
        super().__init__(message)
        self.message = message
        self.data = data


class InvalidDataException(Exception):
    def __init__(self, message, file):
        super().__init__(message)
        self.message = message
        self.file = file


class InvalidMimeTypeException(Exception):
    def __init__(self, message, mime_type):
        super().__init__(message)
        self.message = message + '({})'.format(mime_type)
        self.mime_type = mime_type



#!/usr/bin/env python
# coding: utf-8

import glob
import json
import os
from pprint import pformat

from youtube_dl import YoutubeDL
from youtube_dl.utils import sanitize_filename

from music_dl.core.error import PlaylistParameterException, PlaylistPreprocessException, PlaylistDownloadException
from music_dl.core.youtube_dl import YDLQueueStatus, YDLHelper
from music_dl.lib.log import logger


class Playlist(object):
    """"""""""""""" Initialization """""""""""""""

    def __init__(self, download_url, working_dir, audio_codec='m4a', audio_bitrate=198, playlist_start=1, playlist_end=0, clear_cache=False, verbose=False, test_id=None):
        """
        Initializer

        :param str audio_codec: Preferred audio codec
        :param int audio_bitrate: Preferred audio bitrate
        :param int playlist_start: Index specifying playlist item to start at
        :param int playlist_end: Index specifying playlist item to end at
        :param bool clear_cache: Flag that indicates to delete json file before downloading
        :param bool verbose: Print verbose message
        :param str test_id: Test case identifier
        """

        # Playlist configuration
        self.download_url = download_url
        self.working_dir = working_dir
        self.audio_codec = audio_codec
        self.audio_bitrate = audio_bitrate
        self.playlist_start = playlist_start
        self.playlist_end = playlist_end
        self.clear_cache = clear_cache
        self.verbose = verbose
        self.test_id = test_id

        # Path
        self.download_dir = None
        self.playlist_file = None  # Path to playlist file
        self.downloaded_playlist_file = None  # Path to playlist file contains downloaded songs

        # Playlist data
        self.playlist_data = None  # Playlist data
        self.downloaded_playlist_data = None  # Playlist data
        self.playlist_entry_total = 0  # Total count of entries on playlist
        self.playlist_data_map = {}  # OrderedDict that maps index and entry_id
        self.scheduled_queue_indices = []  # Queue indices to be downloaded
        self.is_playlist = False  # Flag specifies data is playlist

        # Youtube Downloader
        self.ydl = YoutubeDL()
        self.ydl_helper = YDLHelper()

    """"""""""""""" Property """""""""""""""

    @property
    def downloaded_songs_total(self):
        """
        :rtype int
        :return Number of songs downloaded
        """
        if self.downloaded_playlist_data is None:
            return 0
        else:
            return len(self.download_dir.get('entries', []))

    """"""""""""""" Validation """""""""""""""

    def validate(self):
        # Validate playlist start and end
        if self.playlist_start <= 0:
            raise PlaylistParameterException('Invalid start index. Value must be greater than or equal to 1.')

        if 0 < self.playlist_end < self.playlist_start:
            raise PlaylistParameterException('Invalid start and end index. End index must be greater than or equal to start index.')

        # Validate audio codec
        if not (self.audio_codec in ['m4a', 'mp3', 'flac']):
            raise PlaylistParameterException('Supported audio format is m4a, mp3, and flac for now.')

        # Validate audio bit rate
        if self.audio_bitrate < 0:
            raise PlaylistParameterException('Audio bitrate must be positive integer.')

    """"""""""""""" Preprocess """""""""""""""

    def preprocess(self, download_url, working_dir):
        """
        :param str download_url: URL to download
        :param str working_dir: Path to root directory
        """

        self.download_url = download_url

        """ Retrieve playlist """

        logger.info('Retrieving playlist...')
        logger.info('Download URL: {}'.format(self.download_url))

        try:
            ydl_opts = self.ydl_helper.get_preprocess_option(
                download_url=self.download_url,
                audio_codec=self.audio_codec,
                audio_bitrate=self.audio_bitrate,
                playlist_start=self.playlist_start,
                playlist_end=self.playlist_end,
                verbose=self.verbose,
            )
            logger.debug(pformat(ydl_opts))
            self.ydl.__init__(params=ydl_opts)
            # TODO: What is extra_info? Need investigation.
            # self.playlist_data = self.ydl.extract_info(download_url, download=False, process=False, extra_info={})
            self.playlist_data = self.ydl.extract_info(self.download_url, download=False, process=False)

        except:
            raise PlaylistPreprocessException('Could not retrieve playlist.', None)

        if self.playlist_data is None or self.ydl is None:
            raise PlaylistPreprocessException('Could not retrieve playlist.', None)

        logger.info('Done.')

        """ Validate playlist """

        logger.info('Validating playlist...')

        # Determines playlist type
        playlist_extractor = self.playlist_data['extractor'].lower()
        if playlist_extractor == 'youtube:playlist' or playlist_extractor == 'soundcloud:set':
            self.is_playlist = True
            # Define download folder name
            if self.test_id is not None:
                download_folder = self.test_id
            else:
                playlist_title = sanitize_filename(self.playlist_data['title'])
                download_folder = '[{}] {}'.format(self.playlist_data['id'], playlist_title)
            self.download_dir = os.path.join(working_dir, download_folder)

        elif playlist_extractor == 'youtube' or playlist_extractor == 'soundcloud':
            self.is_playlist = False
            # Define download folder name
            if self.test_id is not None:
                download_folder = self.test_id
            else:
                download_folder = 'Single'
            self.download_dir = os.path.join(working_dir, download_folder)

        else:
            raise PlaylistPreprocessException('This playlist is not supported.', self.playlist_data)

        self.playlist_file = os.path.join(self.download_dir, '.queued.json')
        self.downloaded_playlist_file = os.path.join(self.download_dir, '.downloaded.json')

        logger.debug(pformat(self.playlist_data))
        logger.info('Done.')

        """ Create directories """

        logger.info('Creating download directory...')

        # Download directory
        os.makedirs(self.download_dir, exist_ok=True)

        # Playlist
        if self.clear_cache and os.path.exists(self.playlist_file):
            os.remove(self.playlist_file)
        if os.path.exists(self.downloaded_playlist_file):
            os.remove(self.downloaded_playlist_file)

        logger.info('Done.')

        """ Process playlist """

        logger.info('Processing playlist...')

        if self.is_playlist:
            # Convert generator object to list
            self.playlist_data['entries'] = list(self.playlist_data['entries'])
            self.playlist_entry_total = len(self.playlist_data['entries'])

            # Merge playlist
            merged_playlist_data, queue_indices = self.__merge_playlist(self.playlist_data)
            with open(self.playlist_file, 'w') as f:
                json.dump(merged_playlist_data, f, indent=4, ensure_ascii=False)
            self.playlist_data = merged_playlist_data
            self.scheduled_queue_indices = queue_indices

        logger.debug(pformat(self.playlist_data))
        logger.info('Done.')

        return self.download_dir

    """"""""""""""" Download """""""""""""""

    def download(self):
        """
        The function to download songs on the playlist

        :return str is_downloaded: Flag that indicates songs are downloaded
        """

        """ Generate youtube-dl option """

        logger.info('Generating youtube-dl option...')

        ydl_opts = None
        if self.is_playlist and len(self.scheduled_queue_indices) > 0:
            ydl_opts = self.ydl_helper.get_download_option(
                download_dir=self.download_dir,
                hook=self.__download_hook,
                audio_codec=self.audio_codec,
                audio_bitrate=self.audio_bitrate,
                queue_indices=self.scheduled_queue_indices,
                verbose=self.verbose
            )
            logger.debug(pformat(ydl_opts))

        elif not self.is_playlist:
            ydl_opts = self.ydl_helper.get_download_option(
                download_dir=self.download_dir,
                hook=self.__download_hook,
                audio_codec=self.audio_codec,
                audio_bitrate=self.audio_bitrate,
                verbose=self.verbose
            )
            logger.debug(pformat(ydl_opts))

        logger.info('Done.')

        """ Download songs """

        if ydl_opts:
            logger.info('Downloading songs...')

            try:
                # Download playlist
                self.ydl.__init__(params=ydl_opts)
                self.downloaded_playlist_data = self.ydl.extract_info(self.download_url, download=True)
                # Save playlist data
                with open(self.downloaded_playlist_file, 'w') as f:
                    json.dump(self.downloaded_playlist_data, f, indent=4, ensure_ascii=False)

            except Exception:
                raise PlaylistDownloadException('Failed to download playlist.', None)

            if self.downloaded_playlist_data is None:
                raise PlaylistDownloadException('Failed to download playlist.', None)

            logger.info('Done.')
            return True

        else:
            logger.warning('All songs on the playlist are already downloaded. There is nothing to process.')
            return False

    def __download_hook(self, data, queue_index=0, queue_total=0):
        """
        The function that get called on download progress, with a dictionary with the entries.
        More info is available at https://github.com/rg3/youtube-dl/blob/master/youtube_dl/YoutubeDL.py

        :param dict data: Progress information for each queue
        :param int queue_index: Current queue index displayed in log message
        :param int queue_total: Total number of queue displayed in log message
        :return:

        :param dict data: Progress information for each queue
        """

        try:
            # Filename
            song_filename = os.path.basename(data['filename'])
            elapsed = "{0:.2f}".format(data.get('elapsed', -1))

            if queue_index > 0 and queue_total > 0:

                try:
                    # Get index from id-index map
                    song_id, song_ext = os.path.splitext(song_filename)
                    song_index = self.playlist_data_map.get(song_id, -1)
                    entry_found = None
                    if song_index >= 0:
                        # YouTube user playlist or SoundCloud playlist
                        entry_found = self.playlist_data['entries'][song_index]

                    else:
                        # YouTube auto-generated playlist
                        for entry in self.playlist_data['entries']:
                            if entry['id'] == song_id:
                                entry_found = entry
                                break

                    if entry_found:
                        # Finish queue
                        entry_found['status'] = YDLQueueStatus.finished.value
                        with open(self.playlist_file, 'w') as f:
                            json.dump(self.playlist_data, f, indent=4, ensure_ascii=False)

                        # Print song info
                        song_title = self.playlist_data['entries'][song_index].get('title', None)
                        song_title = '[title:{}]'.format(song_title) if song_title else ''
                        elapsed = '[Elapsed:{}]'.format(elapsed) if float(elapsed) >= 0 else ''
                        logger.info('[Process:{}/{}][ID:{}]{}[Size:{}]{} {}'.format(
                            queue_index, queue_total,
                            song_id, song_title, data['_total_bytes_str'], elapsed,
                            'Finished.'
                        ))

                    else:
                        # Print warning
                        song_title = self.playlist_data['entries'][song_index].get('title', None)
                        song_title = '[title:{}]'.format(song_title) if song_title else ''
                        elapsed = '[Elapsed:{}]'.format(elapsed) if float(elapsed) >= 0 else ''
                        logger.warning(
                            '[Process:{}/{}][ID:{}]{}[Size:{}]{} {}'.format(
                                queue_index, queue_total,
                                song_id, song_title, data['_total_bytes_str'], elapsed,
                                'The downloaded song is different from the song on the playlist initially requested. This is caused by YouTube auto-generated playlist.'
                            ))

                except Exception as e:
                    logger.error('[Process:{}/{}] {}:{}'.format(
                        queue_index, queue_total,
                        type(e), str(e),
                    ))

            else:
                song_title, song_ext = os.path.splitext(song_filename)
                elapsed = '[Elapsed:{}]'.format(elapsed) if float(elapsed) >= 0 else ''
                logger.info('[Title:{}][Size:{}]{} {}'.format(
                    song_title, data['_total_bytes_str'], elapsed,
                    'Finished.'
                ))

        except Exception as e:
            logger.error('[Process:{}/{}] {}:{}'.format(queue_index, queue_total, type(e), str(e)))

    def __merge_playlist(self, pl_data):
        """
        The function that merges remote (head_playlist) and local playlist (base_playlist) and generate scheduled queue indices

        :param dict pl_data: Playlist data contains downloaded songs

        :rtype: (dict, list)
        :return: (playlist, indices):  (Merged playlist, Queue indices to download)
        """

        head_playlist_data = pl_data  # Playlist data downloaded from url
        base_playlist_data = None  # Playlist data previously saved on download directory

        """ Load playlist previously saved """

        if os.path.exists(self.playlist_file):
            with open(self.playlist_file) as f:
                # base_playlist_data = json.load(f, object_pairs_hook=OrderedDict)
                base_playlist_data = json.load(f)

        """ Merge Playlist """

        candidate_queue_indices = []
        candidate_queue_index = 1

        # Playlist
        if base_playlist_data:
            # Copy list to avoid index shifting when elements are removed while iterating.
            # https://stackoverflow.com/questions/1207406/how-to-remove-items-from-a-list-while-iterating
            head_index = 0
            head_entries = head_playlist_data['entries'][:]
            for head_entry in head_entries:
                # Delete entry if invalid.
                if head_entry is None or head_entry.get('title', 'N/A').lower() in ['[private video]', '[deleted video]']:
                    song_title = head_entry.get('title', None)
                    song_title = '[title:{}]'.format(song_title) if song_title else ''
                    logger.error('[Playlist:{}/{}][ID:{}]{} {}'.format(
                        head_index + 1, len(head_playlist_data['entries']),
                        head_entry.get('id', 'N/A'), song_title,
                        'The video is private or deleted. Removed from the playlist.'
                    ))
                    del head_playlist_data['entries'][head_index]
                    candidate_queue_index += 1

                else:
                    # Copy list to avoid index shifting when elements are removed while iterating.
                    # https://stackoverflow.com/questions/1207406/how-to-remove-items-from-a-list-while-iterating
                    base_index = 0
                    base_entries = base_playlist_data['entries'][:]
                    for base_entry in base_entries:

                        # Delete entry if invalid.
                        if base_entry is None or base_entry.get('title', 'N/A').lower() in ['[private video]', '[deleted video]']:
                            song_title = base_entry.get('title', None)
                            song_title = '[title:{}]'.format(song_title) if song_title else ''
                            logger.error('[Playlist:{}/{}][ID:{}]{} {}'.format(
                                head_index + 1, len(head_playlist_data['entries']),
                                base_entry.get('id', 'N/A'), song_title,
                                'The video is private or deleted. Removed from the playlist.',
                            ))
                            del base_playlist_data['entries'][base_index]

                        else:
                            # If same entry is found, update status
                            if head_entry['id'] == base_entry['id']:

                                # Merge base status into head status
                                base_entry_status = base_entry.get('status', YDLQueueStatus.ready.value)
                                head_playlist_data['entries'][head_index]['status'] = base_entry_status

                                # Queue index is out of range requested
                                if not self.__is_queue_in_range(head_index):
                                    song_title = base_entry.get('title', None)
                                    song_title = '[title:{}]'.format(song_title) if song_title else ''
                                    logger.debug('[Playlist:{}/{}][ID:{}]{} {}'.format(
                                        head_index + 1, len(head_playlist_data['entries']),
                                        base_entry.get('id', 'N/A'), song_title,
                                        'This queue is out of range requested. Skipped.',
                                    ))

                                # Song is already downloaded
                                elif base_entry_status == YDLQueueStatus.finished.value:
                                    song_title = base_entry.get('title', None)
                                    song_title = '[title:{}]'.format(song_title) if song_title else ''
                                    logger.warning('[Playlist:{}/{}][ID:{}]{} {}'.format(
                                        head_index + 1, len(head_playlist_data['entries']),
                                        base_entry.get('id', 'N/A'), song_title,
                                        'This queue is already finished. Skipped.',
                                    ))

                                # Song is not downloaded yet
                                else:
                                    song_title = base_entry.get('title', None)
                                    song_title = '[title:{}]'.format(song_title) if song_title else ''
                                    logger.info('[Playlist:{}/{}][ID:{}]{} {}'.format(
                                        head_index + 1, len(head_playlist_data['entries']),
                                        base_entry.get('id', 'N/A'), song_title,
                                        'This queue is not finished yet. Added to scheduled queues.',
                                    ))

                                # Delete entry to make iteration faster
                                del base_playlist_data['entries'][base_index]
                                break

                            base_index += 1

                    # Update track number
                    head_entry['track_number'] = head_index + 1

                    # Add queue
                    is_not_finished = head_entry.get('status', YDLQueueStatus.ready.value) != YDLQueueStatus.finished.value
                    if self.__is_queue_in_range(head_index) and is_not_finished:
                        candidate_queue_indices.append(candidate_queue_index)

                    # Add element to dictionary that maps index and entry_id
                    self.playlist_data_map[head_entry['id']] = head_index

                    candidate_queue_index += 1
                    head_index += 1

        # Single song
        else:
            # Copy list to avoid index shifting when elements are removed while iterating.
            # https://stackoverflow.com/questions/1207406/how-to-remove-items-from-a-list-while-iterating
            head_index = 0
            head_entries = head_playlist_data['entries'][:]
            for head_entry in head_entries:
                # Delete entry if invalid.
                if head_entry is None or head_entry.get('title', 'N/A').lower() in ['[private video]', '[deleted video]']:
                    song_title = head_entry.get('title', None)
                    song_title = '[title:{}]'.format(song_title) if song_title else ''
                    logger.error('[Playlist:{}/{}][ID:{}]{} {}'.format(
                        head_index + 1, len(head_playlist_data['entries']),
                        head_entry.get('id', 'N/A'), song_title,
                        'The video is private or deleted. Removed from the playlist.'
                    ))
                    del head_playlist_data['entries'][head_index]
                    candidate_queue_index += 1

                else:
                    song_title = head_entry.get('title', None)
                    song_title = '[title:{}]'.format(song_title) if song_title else ''

                    # Add queue
                    if self.__is_queue_in_range(head_index):
                        candidate_queue_indices.append(candidate_queue_index)
                        logger.info('[Playlist:{}/{}][ID:{}]{} {}'.format(
                            head_index + 1, len(head_playlist_data['entries']),
                            head_entry.get('id', 'N/A'), song_title,
                            'This queue is not finished yet. Added to scheduled queues.'
                        ))
                    else:
                        logger.debug('[Playlist:{}/{}][ID:{}]{} {}'.format(
                            head_index + 1, len(head_playlist_data['entries']),
                            head_entry.get('id', 'N/A'), song_title,
                            'This queue is out of range requested. Skipped.'
                        ))

                    # Update value
                    head_entry['status'] = YDLQueueStatus.ready.value

                    # Update track number
                    head_entry['track_number'] = head_index + 1

                    # Add element to dictionary that maps index and entry_id
                    self.playlist_data_map[head_entry['id']] = head_index

                    candidate_queue_index += 1
                    head_index += 1

        # Save playlist
        with open(self.playlist_file, 'w') as file:
            json.dump(head_playlist_data, file, indent=4, ensure_ascii=False)

        self.playlist_data = head_playlist_data

        return head_playlist_data, candidate_queue_indices

    def __is_queue_in_range(self, index):
        is_greater_than_start = self.playlist_end <= 0 and self.playlist_start <= index + 1
        is_between_start_and_end = self.playlist_start <= index + 1 <= self.playlist_end
        return is_greater_than_start or is_between_start_and_end

    """"""""""""""" Cleanup """""""""""""""

    def cleanup(self):
        """
        The function that deletes all zero byte files from download directory.
        """

        """ Remove zero bytes files created by youtube-dl """

        logger.info('Deleting zero byte files...')
        if not os.path.isdir(self.download_dir):
            raise ValueError('Parameter path is not a directory.')
        files = glob.glob(self.download_dir + '/*')

        # Create generator to find and remove zero byte files
        generator = (os.remove(f) for f in files if os.path.isfile(f) and os.path.getsize(f) == 0)

        try:
            while True:
                next(generator)
        except StopIteration:
            logger.info('Done.')
            return
        except Exception as e:
            logger.error('{}: {}'.format(type(e), e))



#!/usr/bin/env python
# coding: utf-8

import os
import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter, RawTextHelpFormatter, OPTIONAL, SUPPRESS, ZERO_OR_MORE
from gettext import gettext as _

import clipboard
import colorama
import pkg_resources


def parse_args():
    """ Custom Argument Parser """

    class CustomArgumentParser(ArgumentParser):

        def format_help(self):
            return '\n' + super(CustomArgumentParser, self).format_help() + '\n'

        def format_usage(self):
            formatter = self._get_formatter()
            formatter.add_usage(self.usage, self._actions,
                                self._mutually_exclusive_groups)
            return '\n' + formatter.format_help()

        def error(self, message):
            self.print_usage(sys.stderr)
            args = {'prog': self.prog, 'message': message}
            # msg = colorama.Fore.RED + _('\n[%(prog)s] Error: %(message)s\n') % args + colorama.Style.RESET_ALL
            msg = colorama.Fore.RED + _('\nError: %(message)s\n\n') % args + colorama.Style.RESET_ALL
            msg += 'Type "music-dl --help" to show help. \n\n'
            self.exit(2, msg)

    class CustomFormatter(ArgumentDefaultsHelpFormatter, RawTextHelpFormatter):

        def __init__(self, prog, indent_increment=2, max_help_position=40, width=800):
            """
            Initializer that overrides help message indent and positioning
            """
            super().__init__(prog, indent_increment, max_help_position, width)

        def add_usage(self, usage, actions, groups, prefix=None):
            """
            The function that removes prefix from usage
            """
            if prefix is None:
                prefix = ''
                return super(CustomFormatter, self).add_usage(usage, actions, groups, prefix)

        def _format_action_invocation(self, action):
            """
            The function that suppresses duplicated metavars
            https://stackoverflow.com/questions/23936145/python-argparse-help-message-disable-metavar-for-short-options
            """
            if not action.option_strings:
                metavar, = self._metavar_formatter(action, action.dest)(1)
                return metavar
            else:
                parts = []
                # if the Optional doesn't take a value, format is:
                #    -s, --long
                if action.nargs == 0:
                    parts.extend(action.option_strings)

                # if the Optional takes a value, format is:
                #    -s ARGS, --long ARGS
                # change to
                #    -s, --long ARGS
                else:
                    default = action.dest.upper()
                    args_string = self._format_args(action, default)
                    for option_string in action.option_strings:
                        # parts.append('%s %s' % (option_string, args_string))
                        parts.append('%s' % option_string)
                    parts[-1] += ' %s' % args_string
                return ', '.join(parts)

        def _get_help_string(self, action):
            help_msg = action.help
            if '%(default)' not in action.help:
                if action.default is not SUPPRESS:
                    defaulting_nargs = [OPTIONAL, ZERO_OR_MORE]
                    if action.option_strings or action.nargs in defaulting_nargs:
                        if not isinstance(action.default, bool) and action.default is not None:
                            help_msg += ' [Default: %(default)s]'

            return help_msg

    default_dir = os.path.expanduser('~/Music/Downloads')
    pkg_info = pkg_resources.require("music_dl")[0]
    description = "Music Downloader {}".format(pkg_info.version)
    usage = """
Usage: %(prog)s --url http://youtube.com/watch?v=<video_id>&list=<playlist_id>
                --dir ~/Music/Download
                --codec m4a
                --bitrate 128
                --no-track-number
                --no-compilation"""

    """ Parse arguments """

    parser = CustomArgumentParser(
        description=usage,
        usage=description,
        add_help=False,
        formatter_class=CustomFormatter,
    )

    parser._positionals.title = 'Positional Arguments'
    parser._optionals.title = 'Optional Arguments'

    parser.add_argument('-u', '--url', action='store', type=str, metavar='<str>',
                        help='URL to download. [Default: Clipboard Value]')
    parser.add_argument('-d', '--dir', action='store', type=str, metavar='<str>', default=default_dir,
                        help='Path to working directory.')
    parser.add_argument('-c', '--codec', action='store', type=str, metavar='<str> [m4a,mp3,flac]', default='m4a', choices=['m4a', 'mp3', 'flac'],
                        help='Preferred audio codec.')
    parser.add_argument('-b', '--bitrate', action='store', type=int, metavar='<int>', default=198,
                        help='Preferred audio bitrate.')
    parser.add_argument('-s', '--start', action='store', type=int, metavar='<int>', dest='playlist_start', default=1,
                        help='Index specifying playlist item to start at.\nDefault value is index of first song on playlist.')
    parser.add_argument('-e', '--end', action='store', type=int, metavar='<int>', dest='playlist_end', default=0,
                        help='Index specifying playlist item to end at.\nDefault value is index of last song on playlist.')
    parser.add_argument('--no-artwork', action='store_true',
                        help='Forbid adding artwork to audio metadata.')
    parser.add_argument('--no-track-number', action='store_true',
                        help='Forbid adding track number to audio metadata.')
    parser.add_argument('--no-album-title', action='store_true',
                        help='Forbid adding album title to audio metadata.')
    parser.add_argument('--no-album-artist', action='store_true',
                        help='Forbid adding album artist to audio metadata.')
    parser.add_argument('--no-composer', action='store_true',
                        help='Forbid adding composer to audio metadata.')
    parser.add_argument('--no-compilation', action='store_true',
                        help='Forbid adding part of compilation flag to audio metadata.')
    parser.add_argument('--open-dir', action='store_true',
                        help='Open download directory after all songs are downloaded.')
    # parser.add_argument('--clear-cache', action='store_true',
    #                     help='Clear cache directory.')
    parser.add_argument('--verbose', action='store_true',
                        help='Print verbose message.')
    parser.add_argument('--help', action='help', default=SUPPRESS,
                        help='Show this help message and exit.')

    parser_args = parser.parse_args()
    parser_args.url = parser_args.url if parser_args.url is not None else clipboard.paste()
    parser_args.dir = parser_args.dir if parser_args.dir is not None else default_dir

    return parser_args



#!/usr/bin/env python
# coding: utf-8

import logging
import re
import sys

import colorama


class ColorizingStreamHandler(logging.StreamHandler):
    color_map = {
        logging.DEBUG: colorama.Fore.BLUE,
        logging.WARNING: colorama.Fore.YELLOW,
        logging.ERROR: colorama.Fore.RED,
        logging.CRITICAL: colorama.Back.RED,
    }

    def __init__(self, stream, color_map=None):
        logging.StreamHandler.__init__(self, colorama.AnsiToWin32(stream).stream)
        if color_map is not None:
            self.color_map = color_map

    @property
    def is_tty(self):
        isatty = getattr(self.stream, 'isatty', None)
        return isatty and isatty()

    def format(self, record):
        message = logging.StreamHandler.format(self, record)
        if self.is_tty:
            parts = [self.colorize(part, record) for part in message.split('\n', 1)]
            message = '\n'.join(parts)
        else:
            message = self.__remove_ansi(message)
        return message

    def colorize(self, message, record):
        try:
            return (self.color_map[record.levelno] + message +
                    colorama.Style.RESET_ALL)
        except KeyError:
            return message

    def __remove_ansi(self, s):
        ansi_escape = re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
        return ansi_escape.sub('', s)


colorama.init(autoreset=True)
logging.basicConfig(filename='{}.log'.format('music_dl'))
logger = logging.getLogger('music_dl')
logger.setLevel(logging.DEBUG)
handler = ColorizingStreamHandler(sys.stdout)
handler_format = logging.Formatter('[music_dl] %(message)s')
handler.setFormatter(handler_format)
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)



#!/usr/bin/env python
# coding: utf-8

import os
import re
from enum import Enum
from urllib.parse import urlparse

import colorama
from tldextract import tldextract


class YDLHelper(object):
    """"""""""""""" Initializer """""""""""""""

    def __init__(self):
        self.__preprocess_queue_index = 0
        self.__preprocess_queue_total = 0

        self.__download_callback = None
        self.__download_queue_index = 0
        self.__download_queue_total = 0
        self.__download_queue_indices = None

    """"""""""""""" Preprocess """""""""""""""

    def get_preprocess_option(self, download_url, audio_codec='m4a', audio_bitrate=198, playlist_start=1, playlist_end=0, verbose=True):
        """
        The function that creates youtube-dl options.
        More info is available at https://github.com/rg3/youtube-dl/blob/master/youtube_dl/YoutubeDL.py

        :param str download_url: Directory to download songs
        :param str audio_codec: Preferred audio codec
        :param int audio_bitrate: Preferred audio bitrate
        :param int playlist_start: Playlist item to start at
        :param int playlist_end: Playlist item to end at
        :param bool verbose: Print verbose message

        :rtype dict
        :return ydl_opts: Options for youtube-dl
        """
        self.__preprocess_queue_index = playlist_start
        if 0 < playlist_end < playlist_start:
            playlist_end = playlist_start
            self.__preprocess_queue_total = 1
        elif playlist_end <= 0:
            self.__preprocess_queue_total = 0
        else:
            self.__preprocess_queue_index = playlist_end - playlist_start

        ydl_opts = {
            'format': 'bestaudio/best',  # Video format code. See https://github.com/rg3/youtube-dl/blob/master/youtube_dl/options.py for more information
            'writethumbnail': True,  # Write the thumbnail image to a file
            'ignoreerrors': True,  # Do not stop on download errors.
            'nooverwrites': True,  # Prevent overwriting files.
            'forceurl': True,  # Force printing final URL.
            'forcethumbnail': True,  # Force printing thumbnail URL.
            'forcefilename': True,  # Force printing final filename.
            'listformats': False,  # Print an overview of available video formats and exit.
            'postprocessors': [  # More info are available at https://github.com/rg3/youtube-dl/tree/master/youtube_dl/postprocessor
                {
                    'key': 'FFmpegMetadata',
                },
                {
                    'key': 'FFmpegExtractAudio',
                    'preferredcodec': audio_codec,
                    'preferredquality': str(audio_bitrate),
                }
            ],
            'logger': YDLLogger(verbose),
        }

        tld_parsed = tldextract.extract(download_url)
        if tld_parsed.domain == 'youtube':
            url_parsed = urlparse(download_url)
            if ('watch?' in url_parsed and 'v=' in url_parsed and 'list=' in url_parsed) or ('playlist?' in url_parsed and 'list=' in url_parsed):
                ydl_opts['playliststart'] = playlist_start
                ydl_opts['playlistend'] = playlist_end

        elif tld_parsed.domain == 'soundcloud':
            is_match = re.match(r"^(http(s)?://)([\w\.]+)?soundcloud\.com/(.*)/sets/(.*)", download_url)
            if is_match:
                ydl_opts['playliststart'] = playlist_start
                ydl_opts['playlistend'] = playlist_end

        return ydl_opts

    """"""""""""""" Download """""""""""""""

    def get_download_option(self, download_dir, hook, audio_codec='m4a', audio_bitrate=198, queue_indices=None, verbose=True):
        """
        The function that creates youtube-dl options.
        More info is available at https://github.com/rg3/youtube-dl/blob/master/youtube_dl/YoutubeDL.py

        :param str download_dir: Directory to download songs
        :param function hook: Download hook that called when download is finished
        :param list queue_indices: Indices to download
        :param str audio_codec: Preferred audio codec
        :param int audio_bitrate: Preferred audio bitrate
        :param bool verbose: Print verbose message

        :rtype dict
        :return ydl_opts: Options for youtube-dl
        """
        self.__download_callback = hook

        if queue_indices is not None:
            self.__download_queue_indices = queue_indices
            self.__download_queue_index = 1
            self.__download_queue_total = len(queue_indices)

        ydl_opts = {
            # 'outtmpl': Template for output names. See https://github.com/rg3/youtube-dl/blob/master/README.md#output-template for more information
            # 'outtmpl': os.path.join(download_dir, '[%(playlist_index)s-{}] %(title)s.%(ext)s'.format(len(__download_queue_indices))),
            # 'outtmpl': os.path.join(download_dir, '%(title)s.%(ext)s'),
            # 'outtmpl': os.path.join(download_dir, '%(id)s.%(ext)s'),
            'format': 'bestaudio/best',  # Video format code. See https://github.com/rg3/youtube-dl/blob/master/youtube_dl/options.py for more information
            'writethumbnail': True,  # Write the thumbnail image to a file
            'ignoreerrors': True,  # Do not stop on download errors.
            'nooverwrites': True,  # Prevent overwriting files.
            'forceurl': True,  # Force printing final URL.
            'forcethumbnail': True,  # Force printing thumbnail URL.
            'forcefilename': True,  # Force printing final filename.
            'listformats': False,  # Print an overview of available video formats and exit.
            # 'geo_bypass': True,  # Bypass geographic restriction via faking X-Forwarded-For HTTP header.
            'postprocessors': [  # More info are available at https://github.com/rg3/youtube-dl/tree/master/youtube_dl/postprocessor
                # {
                #     'key': 'EmbedThumbnail',  # Embedding thumbnail in m4a is not supported. Using mutagen instead.
                # },
                {
                    'key': 'FFmpegMetadata',
                },
                {
                    'key': 'FFmpegExtractAudio',
                    'preferredcodec': audio_codec,
                    'preferredquality': str(audio_bitrate),
                }
            ],
            'logger': YDLLogger(verbose),
            'progress_hooks': [self.__download_progress_hook],
        }

        # Specify indices of playlist to download
        if self.__download_queue_indices is not None and len(self.__download_queue_indices) > 0:
            ydl_opts['outtmpl'] = os.path.join(download_dir, '%(id)s.%(ext)s')  # Template for output names. See https://github.com/rg3/youtube-dl/blob/master/README.md#output-template for more information
            ydl_opts['playlist_items'] = ','.join([str(index) for index in self.__download_queue_indices])
        else:
            ydl_opts['outtmpl'] = os.path.join(download_dir, '%(title)s.%(ext)s')  # Template for output names. See https://github.com/rg3/youtube-dl/blob/master/README.md#output-template for more information
            pass

        return ydl_opts

    def __download_progress_hook(self, data):
        """
        The function that get called on download progress, with a dictionary with the entries.
        More info is available at https://github.com/rg3/youtube-dl/blob/master/youtube_dl/YoutubeDL.py

        :param dict data: Progress information for each queue
        """
        if data['status'] == 'finished':

            # Playlist
            if self.__download_queue_index > 0:
                # Callback to MusicDL
                self.__download_callback(data, self.__download_queue_index, self.__download_queue_total)
                # Increment queue index
                self.__download_queue_index += 1

            # Single song
            else:
                # Callback to MusicDL
                self.__download_callback(data)


class YDLQueueStatus(Enum):
    ready = 'ready'
    failed = 'failed'
    finished = 'finished'


class YDLLogger(object):
    def __init__(self, verbose=True):
        colorama.init(autoreset=True)
        self.verbose = verbose

    def debug(self, msg):
        if self.verbose:
            print(colorama.Fore.BLUE + msg)

    def info(self, msg):
        print(msg)

    def warning(self, msg):
        print(colorama.Fore.YELLOW + msg)

    def error(self, msg):
        print(colorama.Fore.RED + msg)



#!/usr/bin/env python
# coding: utf-8

import errno
import mimetypes
import os

from mutagen import mp4, id3, flac, File
from youtube_dl.utils import sanitize_filename

from music_dl.core.error import InvalidMimeTypeException, InvalidDataException
from music_dl.lib.log import logger
from music_dl.lib.util import contains_at_least


class MetadataEditor(object):
    """"""""""""""" Initialization """""""""""""""

    def __init__(self, audio_codec, no_artwork, no_album_title, no_album_artist, no_track_number, no_composer, no_compilation, verbose):
        """
        Initializer

        :param str audio_codec: Preferred audio codec
        :param bool no_artwork: Forbid adding artwork to audio metadata
        :param bool no_album_title: Forbid adding track number to audio metadata
        :param bool no_album_artist: Forbid adding album title to audio metadata
        :param bool no_track_number: Forbid adding album artist to audio metadata
        :param bool no_composer: Forbid adding composer to audio metadata
        :param bool no_compilation: Forbid adding part of compilation flag to audio metadata
        :param bool verbose: Print verbose message
        """

        self.audio_codec = audio_codec
        self.no_artwork = no_artwork
        self.no_album_title = no_album_title
        self.no_album_artist = no_album_artist
        self.no_track_number = no_track_number
        self.no_composer = no_composer
        self.no_compilation = no_compilation
        self.verbose = verbose

    """"""""""""""" Update metadata """""""""""""""

    def update(self, download_dir, pl_data, is_playlist):
        """
        The function that update audio metadata.

        :param str download_dir: Download directory
        :param dict pl_data: Playlist data which contains downloaded song information
        :param bool is_playlist: Flag that indicates playlist contains multiple songs
        """

        logger.info('Updating metadata...')

        if is_playlist:
            entries = pl_data.get('entries', [])
            album_title = pl_data.get('title', 'Unknown Album')
            album_artist = pl_data.get('uploader', None)
            album_composer = pl_data.get('extractor_key')

            process_index = 1
            process_total = len(entries)
            for entry in entries:
                try:
                    # Determine filename from entry id
                    song_id = entry['id']
                    # song_index = self.playlist_data_map.get(song_id, -1)

                    song_track_number = process_index
                    song_title = sanitize_filename(entry.get('title', song_id))
                    source_audio_file = os.path.join(download_dir, '{}.{}'.format(song_id, self.audio_codec))

                    # Artwork
                    image_file = None
                    try:
                        image_file = entry['thumbnails'][0]['filename']
                    except:
                        pass

                    # Update tag
                    self.__update_tag(
                        download_dir=download_dir,
                        song_title=song_title,
                        audio_file=source_audio_file,
                        image_file=image_file,
                        album_title=album_title,
                        album_artist=album_artist,
                        album_composer=album_composer,
                        track_number=song_track_number,
                        process_index=process_index,
                        process_total=process_total,
                    )

                except:
                    message = 'Could not update metadata because there is no data found on the playlist. The video may be private or deleted. Audio data is not saved.'
                    logger.error('[Process:{}/{}][Track:{}] {}'.format(process_index, process_total, 'N/A', message))

                process_index += 1

        else:
            base_filename = sanitize_filename(pl_data.get('title', 'Unknown'))
            audio_file = os.path.join(download_dir, '{}.{}'.format(base_filename, self.audio_codec))
            if os.path.exists(audio_file):
                image_file = None
                try:
                    image_file = pl_data['thumbnails'][0]['filename']
                except:
                    pass
                self.__update_tag(
                    download_dir=download_dir,
                    audio_file=audio_file,
                    image_file=image_file
                )

        logger.info('Done.')

    def __update_tag(self, download_dir, audio_file, image_file,
                     song_title=None, album_title=None, album_artist=None, album_composer=None,
                     track_number=-1, process_index=-1, process_total=-1):
        """
        The function that update audio metadata for each song.

        :param str download_dir: Download directory
        :param str audio_file: Path to audio file
        :param str image_file: Path to image file
        :param str song_title: Song title
        :param str album_title: Album title to be saved in metadata
        :param str album_artist: Album artist to be saved in metadata
        :param str album_composer: Album composer to be saved in metadata
        :param int track_number: track number to be saved in metadata
        :param int process_index: Current process index displayed in log message
        :param int process_total: Total number of process displayed in log message
        """

        if audio_file is None:
            logger.warning('[Process:{}/{}][Track:{}] Could not update metadata because there is no data found on the playlist. The video may be private or deleted.'.format(process_index, process_total, track_number))
            return

        if process_index > 0 and process_total > 0:

            if track_number > 0:
                log_prefix = '[Process:{}/{}][Track:{}]'.format(process_index, process_total, track_number)

            else:
                log_prefix = '[Process:{}/{}]'.format(process_index, process_total)

        else:
            log_prefix = ''

        audio_filename = os.path.basename(audio_file)

        try:
            # Validate audio data
            if not os.path.isfile(audio_file):
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), audio_file)

            audio_mime_type = mimetypes.guess_type(audio_file)

            if contains_at_least(audio_mime_type, ['audio/x-mp4', 'audio/x-m4a', 'audio/mp4a-latm']):

                # For more info about mp4 tag is available at
                # https://github.com/quodlibet/mutagen/blob/cf399dc58940fb1356f672809d763be9e2af0033/mutagen/mp4/__init__.py
                # http://atomicparsley.sourceforge.net/mpeg-4files.html
                mp4_data = mp4.MP4(audio_file)
                # Track Number
                if not self.no_track_number and track_number > 0:
                    mp4_data['trkn'] = [(track_number, 0)]
                # Cover image
                if not self.no_artwork:
                    image_data = self.__get_tag_image(image_file, audio_mime_type)
                    if image_data:
                        mp4_data['covr'] = [image_data]
                # Album title
                if not self.no_album_title and album_title is not None:
                    mp4_data['\xa9alb'] = album_title
                # Album artist
                if not self.no_album_artist and album_artist is not None:
                    mp4_data['aART'] = album_artist
                # Composer
                if not self.no_composer and album_composer is not None:
                    mp4_data['\xa9wrt'] = album_composer
                # Part of compilation
                if not self.no_compilation:
                    mp4_data['cpil'] = True
                # Save
                mp4_data.save()

            elif contains_at_least(audio_mime_type, ['audio/x-mp3', 'audio/mpeg']):

                # For more info about ID3v2 tag is available at
                # https://github.com/quodlibet/mutagen/blob/4a5d7d17f1a611280cc52d229aa70b77ca3c55dd/mutagen/id3/_frames.py
                # https://help.mp3tag.de/main_tags.html
                mp3_data = id3.ID3(audio_file)
                # Cover image
                if not self.no_artwork:
                    image_data = self.__get_tag_image(image_file, audio_mime_type)
                    if image_data:
                        mp3_data['APIC'] = image_data
                # Track number
                if not self.no_track_number and track_number > 0:
                    mp3_data.add(id3.TRCK(encoding=3, text=['{}/{}'.format(track_number, 0)]))
                # Album title
                if not self.no_album_title and album_title is not None:
                    mp3_data["TALB"] = id3.TALB(encoding=0, text=album_title)
                # Album artist
                if not self.no_album_artist and album_artist is not None:
                    mp3_data["TPE2"] = id3.TPE2(encoding=0, text=album_artist)
                # Composer
                if not self.no_composer and album_composer is not None:
                    mp3_data["TCOM"] = id3.TCOM(encoding=0, text=album_composer)
                # Part of compilation
                if not self.no_compilation:
                    mp3_data['TCMP'] = id3.TCMP(encoding=0, text=['1'])
                # Save
                mp3_data.save()

            elif contains_at_least(audio_mime_type, ['audio/x-aac']):

                # TODO: Add AAC support
                pass
                # image_data = __get_tag_image(image_file, audio_mime_type)
                # aac_data = aac.AAC(audio_file)
                # if not self.no_track_number:
                #     if track_number > 0 and track_total > 0:
                #         aac_data.add_tags(id3.TRCK(encoding=3, text=['{}/{}'.format(track_number, track_total)]))
                #         # mp3_data['TRCK'] = id3.TRCK(encoding=3, text=[str(track_number)])
                # if image_data:
                #     mp3_data['APIC'] = image_data
                #     aac_data.save()

            elif contains_at_least(audio_mime_type, ['audio/x-flac']):

                # https://github.com/quodlibet/mutagen/blob/a1db79ece62c4e86259f15825e360d1ce0986a22/mutagen/flac.py
                # https://github.com/quodlibet/mutagen/blob/4a5d7d17f1a611280cc52d229aa70b77ca3c55dd/tests/test_flac.py

                flac_data = flac.FLAC(audio_file)
                # Artwork
                if not self.no_artwork:
                    image_data = self.__get_tag_image(image_file, audio_mime_type)
                    if image_data:
                        flac_data.add_picture(image_data)
                # Save
                flac_data.save()

                flac_data = File(audio_file)
                # Track number
                if not self.no_track_number and track_number > 0:
                    flac_data.tags['tracknumber'] = str(track_number)
                # Album title
                if not self.no_album_title and album_title is not None:
                    flac_data.tags['album'] = album_title
                # Album artist
                if not self.no_album_artist and album_artist is not None:
                    flac_data.tags['albumartist'] = album_artist
                # Composer
                if not self.no_composer and album_composer is not None:
                    flac_data.tags['composer'] = album_composer
                # Part of compilation
                if not self.no_compilation:
                    pass
                # Save
                flac_data.save()
                # audio = File(audio_file, easy=True)

            else:
                raise InvalidMimeTypeException("Invalid audio format.", audio_mime_type)

            # Remove artwork if succeeded
            if os.path.exists(image_file):
                os.remove(image_file)

            # Rename filename from id to title
            dest_audio_file = os.path.join(download_dir, '{}.{}'.format(song_title, self.audio_codec))
            os.rename(audio_file, dest_audio_file)

            dest_audio_filename = os.path.basename(dest_audio_file)
            logger.info('{}[File:{}] Updated.'.format(log_prefix, dest_audio_filename))

        except FileNotFoundError:
            message = 'File not found. Skipped.'
            logger.warning('{}[File:{}] {}'.format(log_prefix, audio_filename, message))

        except InvalidDataException as e:
            message = e.message + ' Skipped.'
            logger.warning('{}[File:{}] {}'.format(log_prefix, audio_filename, message))

        except InvalidMimeTypeException as e:
            message = e.message + ' Skipped.'
            logger.warning('{}[File:{}] {}'.format(log_prefix, audio_filename, message))

        except Exception as e:
            message = 'Error {}: {} Skipped.'.format(type(e), str(e))
            logger.error('{}[File:{}] {}'.format(log_prefix, audio_filename, message))

    def __get_tag_image(self, image_file, audio_mime_type):
        """
        The function that returns cover image.

        :param str image_file: Path to image file
        :param list audio_mime_type: Mime-Types of image file
        :return binary image_data: Binary data of image file
        """
        if not os.path.isfile(image_file):
            return

        # Create image
        with open(image_file, 'rb') as f:
            image_data = f.read()
        if image_data is None:
            raise InvalidDataException("Could not create image data.", image_file)

        image_mime_type = mimetypes.guess_type(image_file)[0]

        if contains_at_least(audio_mime_type, ['audio/x-mp4', 'audio/x-m4a', 'audio/mp4a-latm']):

            if image_mime_type == "image/jpeg":
                image_format = mp4.MP4Cover.FORMAT_JPEG
            elif image_mime_type == "image/png":
                image_format = mp4.MP4Cover.FORMAT_PNG
            else:
                raise InvalidMimeTypeException("Invalid image format.", image_mime_type)

            return mp4.MP4Cover(image_data, image_format)

        elif contains_at_least(audio_mime_type, ['audio/x-mp3', 'audio/mpeg']):

            if not image_mime_type == "image/jpeg" and not image_mime_type == "image/png":
                raise InvalidMimeTypeException("Invalid image format.", image_mime_type)

            return id3.APIC(encoding=3, mime=image_mime_type, type=3, desc='Cover', data=image_data)

        elif contains_at_least(audio_mime_type, ['audio/x-aac']):

            # TODO: Add aac support
            pass
            # if not image_mime_type == "image/jpeg" and not image_mime_type == "image/png":
            #     raise InvalidMimeTypeException("Invalid image format.", image_mime_type)
            #
            # picture = flac.Picture()
            # picture.type = 3
            # picture.desc = 'front cover'
            # picture.data = image_data
            #
            # return picture

        elif contains_at_least(audio_mime_type, ['audio/x-flac']):

            if not image_mime_type == "image/jpeg" and not image_mime_type == "image/png":
                raise InvalidMimeTypeException("Invalid image format.", image_mime_type)

            picture = flac.Picture()
            picture.type = 3
            picture.desc = 'front cover'
            picture.data = image_data

            return picture

        else:
            raise InvalidMimeTypeException("Invalid audio format.", audio_mime_type)



#!/usr/bin/env python
# coding: utf-8
import re
from os import path
from pathlib import Path

try:
    from pip._internal.req import parse_requirements
except ImportError:
    from pip.req import parse_requirements

from setuptools import setup, find_packages

root_dir = path.abspath(path.dirname(__file__))

package_name = 'music_dl'


def read_file(filename):
    here = path.abspath(path.dirname(__file__))
    with open(path.join(here, filename), encoding='utf-8') as f:
        return f.read()


def requirements(filename):
    """Parse requirements from requirements.txt."""
    req_file = str(Path(filename))
    req_list = parse_requirements(req_file, session=False)
    return [str(req.req) for req in req_list]


with open(path.join(root_dir, package_name, '__init__.py')) as f:
    init_text = f.read()
    version = re.search(r'__version__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    license = re.search(r'__license__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author = re.search(r'__author__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    author_email = re.search(r'__author_email__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    url = re.search(r'__url__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)
    copyright = re.search(r'__copyright__\s*=\s*[\'\"](.+?)[\'\"]', init_text).group(1)

assert version
assert license
assert author
assert author_email
assert url

setup(
    name=package_name,
    version=version,
    description='Command line tool to download music from YouTube and SoundCloud',
    long_description=read_file('README.md'),
    url=url,
    author=author,
    license=license,
    keywords=['youtube', 'soundcloud', 'download', 'mp3', 'm4a', 'flac', 'youtube-dl', 'youtubedl'],
    author_email=author_email,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.4',
    platforms=['posix'],
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=requirements('requirements.txt'),
    entry_points={
        'console_scripts': [
            'music-dl=music_dl:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/gumob/music-dl/issues',
        'Funding': 'https://donate.pypi.org',
        'Source': 'https://github.com/gumob/music-dl/',
    },
    test_suite='tests',
)



#!/usr/bin/env python
# coding: utf-8

import atexit
import logging
import os
import platform
import subprocess
from urllib.parse import urlparse

import colorama
import pkg_resources
from tldextract import tldextract
from youtube_dl import YoutubeDL

from music_dl.core.error import PlaylistParameterException, PlaylistPreprocessException, DirectoryException
from music_dl.core.metadata import MetadataEditor
from music_dl.core.playlist import Playlist
from music_dl.core.youtube_dl import YDLHelper
from music_dl.lib.dir import is_path_exists_or_creatable
from music_dl.lib.log import logger


class MusicDL(object):
    """"""""""""""" Initializer """""""""""""""

    def __init__(self,
                 download_url,
                 working_dir=os.path.expanduser('~/Music/Downloads'),
                 audio_codec='m4a',
                 audio_bitrate=198,
                 playlist_start=1,
                 playlist_end=0,
                 no_artwork=False,
                 no_album_title=False,
                 no_album_artist=False,
                 no_track_number=False,
                 no_composer=False,
                 no_compilation=False,
                 open_dir=False,
                 verbose=False,
                 clear_cache=False,
                 test_id=None):
        # Download URL
        self.download_url = download_url
        # Working directory
        self.working_dir = working_dir
        # Open directory configuration
        self.open_dir = open_dir
        # Verbose
        self.verbose = verbose
        # Music Downloader
        self.playlist = Playlist(
            download_url=download_url,
            working_dir=working_dir,
            audio_codec=audio_codec,
            audio_bitrate=audio_bitrate,
            playlist_start=playlist_start,
            playlist_end=playlist_end,
            clear_cache=clear_cache,
            verbose=verbose,
            test_id=test_id,
        )
        self.metadata_editor = MetadataEditor(
            audio_codec=audio_codec,
            no_artwork=no_artwork,
            no_album_title=no_album_title,
            no_album_artist=no_album_artist,
            no_track_number=no_track_number,
            no_composer=no_composer,
            no_compilation=no_compilation,
            verbose=verbose
        )
        # Youtube Downloader
        self.ydl = YoutubeDL()
        self.ydl_helper = YDLHelper()

    """"""""""""""" Download Music (Public) """""""""""""""

    def download(self):
        """
        The function that downloads songs from YouTube and SoundCloud

        :return bool result: Result of process, used by unit test
        """
        print()
        atexit.register(print)

        """ Set log level """

        if self.verbose:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

        """ Print version """

        logger.info(pkg_resources.require("music_dl")[0])

        """ Validate parameters """

        logger.info('Validating parameters...')

        try:
            # Validate download url
            url_parsed = urlparse(self.download_url)
            if not url_parsed.scheme.startswith('http'):
                raise DirectoryException('Invalid URL. URL must start with http*. Input value is {}'.format(self.download_url))
            tld_parsed = tldextract.extract(self.download_url)
            if not (tld_parsed.domain in ['youtube', 'soundcloud']):
                raise DirectoryException('Invalid URL. Music Downloader supports only YouTube and SoundCloud. Input value is {}'.format(self.download_url))
            # Validate download directory
            if not is_path_exists_or_creatable(self.working_dir):
                raise DirectoryException('Invalid directory. Please specify valid download directory. Input value is {}'.format(self.working_dir))

        except DirectoryException as e:
            logger.error(e.message)
            logger.fatal('Aborted.')
            exit()

        # Validate playlist configuration
        try:
            self.playlist.validate()

        except PlaylistParameterException as e:
            logger.error(e.message)
            logger.fatal('Aborted.')
            exit()

        logger.info('Done.')

        """ Retrieve playlist """

        download_dir = None
        try:
            download_dir = self.playlist.preprocess(self.download_url, self.working_dir)

        except PlaylistPreprocessException as e:
            logger.error(e.message)
            logger.error(e.data)
            logger.fatal('Aborted.')
            exit()

        """ Download playlist """

        is_downloaded = False
        try:
            is_downloaded = self.playlist.download()

        except PlaylistPreprocessException as e:
            logger.error(e.message)
            logger.error(e.data)
            logger.fatal('Aborted.')
            exit()

        """ Update metadata """

        if is_downloaded:
            self.metadata_editor.update(
                download_dir=self.playlist.download_dir,
                pl_data=self.playlist.downloaded_playlist_data,
                is_playlist=self.playlist.is_playlist,
            )

        """ Cleanup download directory """

        self.playlist.cleanup()

        """ Print completion message """

        logger.info('All process has done.')
        logger.info('Now you can find downloaded songs at {}'.format(colorama.Fore.LIGHTCYAN_EX + download_dir))

        """ Open download directory """

        if self.open_dir and platform.system().lower() == 'darwin':
            subprocess.check_output(['open', download_dir], stderr=subprocess.STDOUT)

        return True
